# Copyright (c) 2022 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
from pathlib import Path
from typing import Any, Callable, List, Mapping, Optional, Type, Union

from hydra._internal.callbacks import Callbacks
from hydra._internal.hydra import Hydra
from hydra._internal.utils import create_config_search_path
from hydra.core.config_store import ConfigStore
from hydra.core.global_hydra import GlobalHydra
from hydra.core.utils import JobReturn
from hydra.plugins.sweeper import Sweeper
from hydra.types import HydraContext, RunMode
from omegaconf import DictConfig, OmegaConf

from hydra_zen._hydra_overloads import instantiate
from hydra_zen.typing._implementations import DataClass


def _store_config(
    cfg: Union[DataClass, Type[DataClass], DictConfig, Mapping],
    config_name: str = "hydra_launch",
) -> str:
    """Stores configuration object in Hydra's ConfigStore.

    Parameters
    ----------
    cfg : Union[DataClass, DictConfig, Mapping]
        A configuration as a dataclass, configuration object, or a dictionary.

    config_name : str (default: hydra_launch)
        The configuration name used to store the configuration.

    Returns
    -------
    config_name : str
        The configuration name used to store the default configuration.

    Notes
    -----
    The input configuration is registered in the Hydra ConfigStore [1]_ using a
    user-provided config name.

    References
    ----------
    .. [1] https://hydra.cc/docs/tutorials/structured_config/config_store
    """
    cs = ConfigStore().instance()
    cs.store(name=config_name, node=cfg)
    return config_name


def launch(
    config: Union[DataClass, Type[DataClass], Mapping[str, Any]],
    task_function: Callable[[DictConfig], Any],
    overrides: Optional[List[str]] = None,
    config_dir: Optional[Union[str, Path]] = None,
    config_name: str = "zen_launch",
    job_name: str = "zen_launch",
    with_log_configuration: bool = True,
    multirun: bool = False,
) -> Union[JobReturn, Any]:
    r"""Launch a Hydra job using a Python-based interface.

    `launch` is designed to closely match the interface of the standard Hydra CLI.
    For example, launching a Hydra job from the CLI via::

       $ python my_task.py job/group=group_name job.group.param=1

    corresponds to the following usage of `launch`:

       >>> job = launch(config, task_function, overrides=["job/group=group_name", "job.group.param=1"])

    Parameters
    ----------
    config : DataClass | Type[DataClass] | Mapping[str, Any]
        A config that will be passed to ``task_function``.

    task_function : Callable[[DictConfig], Any]
        The function that Hydra will execute. Its input will be ``config``, which
        has been modified via the specified ``overrides``

    overrides : Optional[List[str]]
        If provided, sets/overrides values in ``config``. See [1]_ and [2]_
        for a detailed discussion of the "grammar" supported by ``overrides``.

    config_dir : Optional[Union[str, Path]]
        The config path: a directory where Hydra will search for configs.

    config_name : str (default: "zen_launch")
        Name of the stored configuration in Hydra's ConfigStore API.

    job_name : str (default: "zen_launch")

    with_log_configuration : bool (default: True)
        If ``True``, enables the configuration of the logging subsystem from the loaded config.

    multirun : bool (default: False)
        Launch a Hydra multi-run ([3]_).

    Returns
    -------
    result : JobReturn | Any
        If ``multirun is False``:
            A ``JobReturn`` object storing the results of the Hydra experiment via the following attributes
                - ``cfg``: Reflects ``config``
                - ``overrides``: Reflects ``overrides``
                - ``return_value``: The return value of the task function
                - ``hydra_cfg``: The Hydra configuration object
                - ``working_dir``: The experiment working directory
                - ``task_name``: The task name of the Hydra job
                - ``status``: A ``JobStatus`` enum reporting whether or not the job completed successfully
        Else:
            Return values of all launched jobs (depends on the Sweeper implementation).

    References
    ----------
    .. [1] https://hydra.cc/docs/next/advanced/override_grammar/basic
    .. [2] https://hydra.cc/docs/next/configure_hydra/intro
    .. [3] https://hydra.cc/docs/tutorials/basic/running_your_app/multi-run

    Examples
    --------

    **Basic usage**

    Let's define and launch a trivial Hydra app.

    >>> from hydra_zen import make_config, launch, to_yaml

    First, we will define a config, which determines the configurable interface to our
    "app". For the purpose of example, we'll design the "interface" of this config to accept
    two configurable parameters: ``a`` and ``b``.

    >>> Conf = make_config("a", "b")

    Our task function accepts the config as an input and uses it to run some generic functionality.
    For simplicity's sake, let's design this task function to: convert the job's config to a
    yaml-formatted string, print it, and then return the string.

    >>> def task_fn(cfg):
    ...     out = to_yaml(cfg)  # task's input config, converted to yaml-string
    ...     print(out)
    ...     return out

    Now, let's use `launch` to run this task function via Hydra, using particular configured
    values (or, "overrides") for ``a`` and ``b``.

    >>> job_out = launch(Conf, task_fn, overrides=["a=1", "b='foo'"])
    a: 1
    b: foo

    Let's inspect ``job_out`` to see the ways that it summarizes the results of this job.

    >>> job_out.return_value  # the value returned by `task_fn`
    'a: 1\nb: foo\n'

    >>> job_out.working_dir  # where the job's outputs, logs, and configs are saved
    'outputs/2021-10-19/15-27-11'

    >>> job_out.cfg  # the particular config used to run our task-function
    {'a': 1, 'b': 'foo'}

    >>> job_out.overrides  # the overrides that we provides
    ['a=1', "b='foo'"]

    >>> job_out.status  # the job's completion status
    <JobStatus.COMPLETED: 1>

    **Launching a multirun job**

    We can launch multiple runs of our task-function, using various configured values.
    Let's launch a multirun that sweeps over three configurations

    >>> (outputs,) = launch(
    ...     Conf,
    ...     task_fn,
    ...     overrides=["a=1,2,3", "b='bar'"],
    ...     multirun=True,
    ... )
    [2021-10-19 17:50:07,334][HYDRA] Launching 3 jobs locally
    [2021-10-19 17:50:07,334][HYDRA] 	#0 : a=1 b='bar'
    a: 1
    b: bar
    [2021-10-19 17:50:07,434][HYDRA] 	#1 : a=2 b='bar'
    a: 2
    b: bar
    [2021-10-19 17:50:07,535][HYDRA] 	#2 : a=3 b='bar'
    a: 3
    b: bar

    ``outputs`` contains three corresponding ``JobReturns`` instances.

    >>> len(outputs)
    3
    >>> [j.cfg for j in outputs]
    [{'a': 1, 'b': 'bar'}, {'a': 2, 'b': 'bar'}, {'a': 3, 'b': 'bar'}]

    Each run's outputs, logs, and configs are saved to separate working directories

    >>> [j.working_dir for j in outputs]
    ['multirun/2021-10-19/17-50-07\\0',
    'multirun/2021-10-19/17-50-07\\1',
    'multirun/2021-10-19/17-50-07\\2']
    """
    config_name = _store_config(config, config_name)

    if config_dir is not None:
        config_dir = str(Path(config_dir).absolute())
    search_path = create_config_search_path(config_dir)

    hydra = Hydra.create_main_hydra2(task_name=job_name, config_search_path=search_path)
    try:
        if not multirun:
            # Here we can use Hydra's `run` method
            job = hydra.run(
                config_name=config_name,
                task_function=task_function,
                overrides=overrides if overrides is not None else [],
                with_log_configuration=with_log_configuration,
            )

        else:
            # Instead of running Hydra's `multirun` method we instantiate
            # and run the sweeper method.  This allows us to run local
            # sweepers and launchers without installing them in `hydra_plugins`
            # package directory.
            cfg = hydra.compose_config(
                config_name=config_name,
                overrides=overrides if overrides is not None else [],
                with_log_configuration=with_log_configuration,
                run_mode=RunMode.MULTIRUN,
            )

            callbacks = Callbacks(cfg)
            callbacks.on_multirun_start(config=cfg, config_name=config_name)

            # Instantiate sweeper without using Hydra's Plugin discovery (Zen!)
            sweeper = instantiate(cfg.hydra.sweeper)
            assert isinstance(sweeper, Sweeper)
            sweeper.setup(
                config=cfg,
                hydra_context=HydraContext(
                    config_loader=hydra.config_loader, callbacks=callbacks
                ),
                task_function=task_function,
            )

            task_overrides = OmegaConf.to_container(
                cfg.hydra.overrides.task, resolve=False
            )
            assert isinstance(task_overrides, list)

            job = sweeper.sweep(arguments=task_overrides)
            callbacks.on_multirun_end(config=cfg, config_name=config_name)

    finally:
        GlobalHydra.instance().clear()
    return job



# Copyright (c) 2022 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

from ._implementations import (
    Builds,
    Importable,
    Just,
    Partial,
    PartialBuilds,
    SupportedPrimitive,
)

__all__ = [
    "Builds",
    "Importable",
    "Just",
    "Partial",
    "PartialBuilds",
    "SupportedPrimitive",
]



# Copyright (c) 2022 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

import warnings
from pathlib import Path
from typing import Any, Callable, List, Mapping, Optional, Union

from omegaconf.dictconfig import DictConfig

from hydra_zen.errors import HydraZenDeprecationWarning
from hydra_zen.typing._implementations import DataClass

from ._launch import launch


def hydra_run(
    config: Union[DataClass, DictConfig, Mapping],
    task_function: Callable[[DictConfig], Any],
    overrides: Optional[List[str]] = None,
    config_dir: Optional[Union[str, Path]] = None,
    config_name: str = "hydra_run",
    job_name: str = "hydra_run",
    with_log_configuration: bool = True,
):
    """(Deprecated) Launch a Hydra job defined by `task_function` using the configuration
    provided in `config`.

    .. deprecated:: 0.3.0
          `hydra_run` will be removed in hydra-zen 1.0.0; it is replaced by
          :func:`hydra_zen.launch`.

    Parameters
    ----------
    config : Union[DataClass, DictConfig, Mapping]
        A configuration as a dataclass, configuration object, or a dictionary.

    task_function : Callable[[DictConfig], Any]
        The function Hydra will execute with the given configuration.

    overrides : Optional[List[str]] (default: None)
        If provided, overrides default configurations, see [1]_ and [2]_.

    config_dir : Optional[Union[str, Path]] (default: None)
        Add configuration directories if needed.

    config_name : str (default: "hydra_run")
        Name of the stored configuration in Hydra's ConfigStore API.

    job_name : str (default: "hydra_run")

    with_log_configuration : bool (default: True)
        Flag to configure logging subsystem from the loaded config

    Returns
    -------
    result : JobReturn

    References
    ----------
    .. [1] https://hydra.cc/docs/next/advanced/override_grammar/basic
    .. [2] https://hydra.cc/docs/next/configure_hydra/intro
    .. [3] https://hydra.cc/docs/tutorials/basic/running_your_app/multi-run
    """
    warnings.warn(
        HydraZenDeprecationWarning(
            "hydra_zen.experimental.hydra_run is deprecated "
            "as of 2021-10-27. Change `hydra_run(cfg, task_fn, overrides, ...)` to `launch(cfg, task_fn, overrides, ...)`."
            "\n\nThis will be an error in hydra-zen 1.0.0, or by 2022-01-27 — whichever "
            "comes first.\n\nNote: This deprecation does not impact yaml configs "
            "produced by `builds`."
        ),
        stacklevel=2,
    )
    return launch(
        config,
        task_function=task_function,
        overrides=overrides,
        config_dir=config_dir,
        config_name=config_name,
        job_name=job_name,
        with_log_configuration=with_log_configuration,
    )


def hydra_multirun(
    config: Union[DataClass, DictConfig, Mapping],
    task_function: Callable[[DictConfig], Any],
    overrides: Optional[List[str]] = None,
    config_dir: Optional[Union[str, Path]] = None,
    config_name: str = "hydra_run",
    job_name: str = "hydra_run",
    with_log_configuration: bool = True,
):
    """(Deprecated) Launch multiple Hydra jobs defined by `task_function` using the configuration
    provided in `config`.

    .. deprecated:: 0.3.0
          `hydra_multirun` will be removed in hydra-zen 1.0.0; it is replaced by
          :func:`hydra_zen.launch`.

    Parameters
    ----------
    config : Union[DataClass, DictConfig, Mapping]
        A configuration as a dataclass, configuration object, or a dictionary.

    task_function : Callable[[DictConfig], Any]
        The function Hydra will execute with the given configuration.

    overrides : Optional[List[str]] (default: None)
        If provided, overrides default configurations, see [1]_ and [2]_.

    config_dir : Optional[Union[str, Path]] (default: None)
        Add configuration directories if needed.

    config_name : str (default: "hydra_run")
        Name of the stored configuration in Hydra's ConfigStore API.

    job_name : str (default: "hydra_run")

    with_log_configuration : bool (default: True)
        Flag to configure logging subsystem from the loaded config

    Returns
    -------
    result : Any

    References
    ----------
    .. [1] https://hydra.cc/docs/next/advanced/override_grammar/basic
    .. [2] https://hydra.cc/docs/next/configure_hydra/intro
    .. [3] https://hydra.cc/docs/tutorials/basic/running_your_app/multi-run
    """
    warnings.warn(
        HydraZenDeprecationWarning(
            "hydra_zen.experimental.hydra_multirun is deprecated "
            "as of 2021-10-27. Change `hydra_multirun(cfg, task_fn, overrides, ...)` to `launch(cfg, task_fn, overrides, multirun=True, ...)`."
            "\n\nThis will be an error in hydra-zen 1.0.0, or by 2022-01-27 — whichever "
            "comes first.\n\nNote: This deprecation does not impact yaml configs "
            "produced by `builds`."
        ),
        stacklevel=2,
    )
    return launch(
        config,
        task_function=task_function,
        overrides=overrides,
        config_dir=config_dir,
        config_name=config_name,
        job_name=job_name,
        with_log_configuration=with_log_configuration,
        multirun=True,
    )



# Copyright (c) 2022 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

from dataclasses import Field
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    Mapping,
    NewType,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from omegaconf import DictConfig, ListConfig
from typing_extensions import Protocol, TypedDict, runtime_checkable

__all__ = [
    "Just",
    "Builds",
    "PartialBuilds",
    "Partial",
    "Importable",
    "SupportedPrimitive",
]


class EmptyDict(TypedDict):
    pass


_T = TypeVar("_T", covariant=True)
T2 = TypeVar("T2")
T3 = TypeVar("T3")


@runtime_checkable
class Partial(Protocol[T2]):
    func: Callable[..., T2]
    args: Tuple[Any, ...]
    keywords: Dict[str, Any]

    def __new__(
        cls: Type[T3], func: Callable[..., T2], *args: Any, **kwargs: Any
    ) -> T3:  # pragma: no cover
        ...

    def __call__(self, *args: Any, **kwargs: Any) -> T2:  # pragma: no cover
        ...


InterpStr = NewType("InterpStr", str)


class _DataClass(Protocol):  # pragma: no cover
    # doesn't provide __init__, __getattribute__, etc.
    __dataclass_fields__: Dict[str, Field]


class DataClass(_DataClass, Protocol):  # pragma: no cover
    def __init__(self, *args, **kwargs) -> None:
        ...

    def __getattribute__(self, __name: str) -> Any:
        ...

    def __setattr__(self, __name: str, __value: Any) -> None:
        ...


@runtime_checkable
class Builds(DataClass, Protocol[_T]):  # pragma: no cover
    _target_: str


@runtime_checkable
class Just(Builds, Protocol[_T]):  # pragma: no cover
    path: str  # interpolated string for importing obj
    _target_: str = "hydra_zen.funcs.get_obj"


@runtime_checkable
class PartialBuilds(Builds, Protocol[_T]):  # pragma: no cover
    _target_: str = "hydra_zen.funcs.zen_processing"
    _zen_target: str
    _zen_partial: bool = True


@runtime_checkable
class HydraPartialBuilds(Builds, Protocol[_T]):  # pragma: no cover
    _partial_: bool = True


@runtime_checkable
class HasTarget(Protocol):  # pragma: no cover
    _target_: str


@runtime_checkable
class HasPartialTarget(Protocol):  # pragma: no cover
    _zen_partial: bool = True


Importable = TypeVar("Importable", bound=Callable)

_HydraPrimitive = Union[
    bool,
    None,
    int,
    float,
    str,
]

_SupportedPrimitive = Union[
    _HydraPrimitive,
    ListConfig,
    DictConfig,
    type,
    Callable,
    Enum,
    _DataClass,
    complex,
    Path,
    Partial,
    range,
    set,
    EmptyDict,  # not covered by Mapping[..., ...]
]

SupportedPrimitive = Union[
    _SupportedPrimitive,
    FrozenSet["SupportedPrimitive"],
    # Even thought this is redundant with Sequence, it seems to
    # be needed for pyright to do proper checking of tuple contents
    Tuple["SupportedPrimitive"],
    # Mutable generic containers need to be invariant, so
    # we have to settle for Sequence/Mapping. While this
    # is overly permissive in terms of sequence-type, it
    # at least affords quality checking of sequence content
    Sequence["SupportedPrimitive"],
    # Mapping is covariant only in value
    Mapping[Any, "SupportedPrimitive"],
]



# Copyright (c) 2022 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

"""
Simple helper functions used to implement `just` and `builds`. This module is designed specifically so
that these functions have a legible module-path when they appear in configuration files.
"""
import functools as _functools
import typing as _typing

from hydra._internal import utils as _hydra_internal_utils
from hydra.utils import log as _log

from hydra_zen.structured_configs._utils import (
    is_interpolated_string as _is_interpolated_string,
)
from hydra_zen.typing import Partial as _Partial

__all__ = ["partial", "get_obj", "zen_processing"]

_T = _typing.TypeVar("_T")
_WrapperConf = _typing.Union[
    str, _typing.Callable[[_typing.Callable], _typing.Callable]
]


def partial(
    *args: _typing.Any,
    _partial_target_: _typing.Callable[..., _T],
    **kwargs: _typing.Any,
) -> _Partial[_T]:
    """Equivalent to ``functools.partial`` but provides a named parameter for the callable."""
    return _typing.cast(
        _Partial[_T], _functools.partial(_partial_target_, *args, **kwargs)
    )


def get_obj(*, path: str) -> _typing.Union[type, _typing.Callable[..., _typing.Any]]:
    """Imports an object given the specified path."""
    try:
        cl = _hydra_internal_utils._locate(path)
        return cl
    except Exception as e:  # pragma: no cover
        _log.error(f"Error getting callable at {path} : {e}")
        raise e


def zen_processing(
    *args,
    _zen_target: str,
    _zen_partial: bool = False,
    _zen_exclude: _typing.Sequence[str] = tuple(),
    _zen_wrappers: _typing.Union[
        _WrapperConf, _typing.Sequence[_WrapperConf]
    ] = tuple(),
    **kwargs,
):
    if isinstance(_zen_wrappers, str) or not isinstance(
        _zen_wrappers, _typing.Sequence
    ):
        unresolved_wrappers: _typing.Sequence[_WrapperConf] = (_zen_wrappers,)  # type: ignore
    else:
        unresolved_wrappers: _typing.Sequence[_WrapperConf] = _zen_wrappers
    del _zen_wrappers

    resolved_wrappers = []

    for _unresolved in unresolved_wrappers:
        if _unresolved is None:
            # We permit interpolated fields to resolve to `None`; this is
            # a nice pattern for enabling people to ergonomically toggle
            # wrappers off.
            continue
        if isinstance(_unresolved, str):
            # Hydra will have already raised on missing interpolation
            # keys by here
            assert not _is_interpolated_string(_unresolved)
            _unresolved = get_obj(path=_unresolved)

        if not callable(_unresolved):
            raise TypeError(
                f"Instantiating {_zen_target}: `zen_wrappers` was passed a non-callable object: {_unresolved}"
            )
        else:
            resolved = _unresolved
        del _unresolved
        resolved_wrappers.append(resolved)

    obj = get_obj(path=_zen_target)

    # first wrapper listed should be called first
    # [f1, f2, f3, ...] ->
    #    target = f1(target)
    #    target = f2(target)
    #    target = f3(target)
    #    ...
    for wrapper in resolved_wrappers:
        obj = wrapper(obj)

    if _zen_exclude:
        excluded_set = set(_zen_exclude)
        kwargs = {k: v for k, v in kwargs.items() if k not in excluded_set}

    if _zen_partial is True:
        return _functools.partial(obj, *args, **kwargs)
    return obj(*args, **kwargs)



# Copyright (c) 2022 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

from ._implementations import (
    ZenField,
    builds,
    hydrated_dataclass,
    just,
    make_config,
    make_custom_builds_fn,
    mutable_value,
)

__all__ = [
    "builds",
    "just",
    "hydrated_dataclass",
    "mutable_value",
    "make_config",
    "ZenField",
    "make_custom_builds_fn",
]



# Copyright (c) 2022 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

from setuptools import find_packages, setup

import versioneer

DISTNAME = "hydra_zen"
LICENSE = "MIT"
AUTHOR = "Justin Goodwin, Ryan Soklaski"
AUTHOR_EMAIL = "ryan.soklaski@ll.mit.edu"
URL = "https://github.com/mit-ll-responsible-ai/hydra-zen"
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Intended Audience :: Science/Research",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Topic :: Scientific/Engineering",
]
KEYWORDS = (
    "machine learning research configuration scalable reproducible yaml Hydra dataclass"
)
INSTALL_REQUIRES = [
    "hydra-core >= 1.1.0",
    "typing-extensions >= 3.10.0.1",
]
TESTS_REQUIRE = [
    "pytest >= 3.8",
    "hypothesis >= 6.28.0",
]

DESCRIPTION = "Configurable, reproducible, and scalable workflows in Python, via Hydra"
LONG_DESCRIPTION = """
hydra-zen is a Python library that simplifies the process of writing code (research-grade or production-grade) that is:

- **Configurable**: you can configure all aspects of your code from a single interface (the command line or a single Python function).
- **Repeatable**: each run of your code will be self-documenting; the full configuration of your software is saved alongside your results.
- **Scalable**: launch multiple runs of your software, be it on your local machine or across multiple nodes on a cluster.

It builds off – and is fully compatible with – Hydra, a framework for elegantly
 configuring complex applications.

hydra-zen helps simplify the process of using Hydra by providing convenient functions
for creating and validating configs, as well as launching Hydra jobs. It also provides
novel functionality such as wrapped instantiation and meta fields in configs.
"""


setup(
    name=DISTNAME,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    license=LICENSE,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    classifiers=CLASSIFIERS,
    keywords=KEYWORDS,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRE,
    url=URL,
    download_url="https://github.com/mit-ll-responsible-ai/hydra-zen/tarball/v"
    + versioneer.get_version(),
    python_requires=">=3.6",
    packages=find_packages(where="src", exclude=["tests", "tests.*"]),
    package_dir={"": "src"},
    package_data={"hydra_zen": ["py.typed"]},
    extras_require={
        "pydantic": [
            "pydantic>=1.8.2"
        ],  # don't reduce below 1.8.2 -- security vulnerability
        "beartype": ["beartype>=0.8.0"],
    },
)



import sys

if sys.version_info < (3, 7):  # pragma: no cover
    raise NotImplementedError(
        "Features that utilize `hydra_zen._utils.coerce` "
        "(e.g. beartype-validation) require Python 3.7 or greater."
    )

import inspect
from collections import deque
from functools import wraps
from typing import (
    Callable,
    NamedTuple,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
    get_type_hints,
)

from omegaconf import ListConfig
from typing_extensions import TypeGuard

from hydra_zen.structured_configs._utils import get_args, get_origin

_T = TypeVar("_T", bound=Callable)

__all__ = ["coerce_sequences"]


def _is_namedtuple_type(x) -> TypeGuard[Type[NamedTuple]]:  # pragma: no cover

    try:
        bases = x.__bases__
        fields = x._fields
    except AttributeError:
        return False

    if bases is None:
        return False

    if len(bases) != 1 or bases[0] is not tuple:
        return False

    if not isinstance(fields, tuple):
        return False

    return all(isinstance(fieldname, str) for fieldname in fields)


_NoneType = type(None)

_POSITIONAL_ONLY = inspect.Parameter.POSITIONAL_ONLY
_VAR_POSITIONAL = inspect.Parameter.VAR_POSITIONAL
_KEYWORD_ONLY = inspect.Parameter.KEYWORD_ONLY
_VAR_KEYWORD = inspect.Parameter.VAR_KEYWORD


def coerce_sequences(obj: _T) -> _T:
    """
    Hydra is only able to read non-string sequences as lists (or ListConfig).
    This wrapper will cast these lists to their desired type, based
    on the annotated-type associated with that sequence.

    This is a no-op for objects that don't have any annotated non-string
    sequence types.

    This is strictly an experimental utility. Use at your own risk.

    Parameters
    ----------
    obj : Callable

    Returns
    -------
    wrapped_obj

    Notes
    -----
    The supported non-string sequence types are:
    - tuples
    - deques
    - lists
    - named-tuples

    The only Union-based types that are supported are of the form
    ``Optional[<sequence-type>]``.

    Examples
    --------
    >>> from hydra_zen import builds, instantiate
    >>> def f(x: tuple): return x

    Without wrapping:

    >>> conf_no_wrap = builds(f, x=(1, 2))
    >>> instantiate(conf_no_wrap)
    [1, 2]

    With wrapping:

    >>> conf_wrapped = builds(f, x=(1, 2), zen_wrappers=convert_sequences)
    >>> instantiate(conf_wrapped)
    (1, 2)
    """
    if inspect.isclass(obj) and hasattr(type, "__init__"):
        hints = get_type_hints(obj.__init__)
    else:
        hints = get_type_hints(obj)

    if not hints:
        return cast(_T, obj)

    # We need to keep track of parameters that are annotated with
    # sequence-like annotations
    list_caster_by_pos = {}  # by pos-index in the signature
    list_caster_by_name = {}  # by param-name in the signature

    sig = inspect.signature(obj)
    for n, (name, param) in enumerate(sig.parameters.items()):
        if name not in hints or param.kind in {_VAR_KEYWORD, _VAR_POSITIONAL}:
            continue

        annotation = hints[name]

        _origin = get_origin(annotation)

        if _origin is not None and _origin is Union:
            # Check for Optional[A]
            _args = get_args(annotation)
            if len(_args) == 2:
                if _args[0] is _NoneType:
                    annotation = _args[1]
                elif _args[1] is _NoneType:
                    annotation = _args[0]
                else:  # pragma: no cover
                    # E.g. Union[A, B]
                    # we cover this  in tests coverage-tooling struggles here
                    continue
            else:  # pragma: no cover
                # E.g. Union[A, B, C]
                # we cover this in tests but coverage-tooling struggles here
                continue
        del _origin

        if not isinstance(annotation, type):
            # E.g. annotation = Tuple[int, int]
            # Tuple[int, int].__origin__ -> tuple
            caster = get_origin(annotation)
            if caster is None:  # pragma: no cover
                continue
        else:
            # E.g. annotation = tuple
            caster = annotation

        if (
            inspect.isclass(caster)
            and issubclass(caster, Sequence)
            # and not inspect.isabstract(caster)  # E.g. caster = Sequence
            and not issubclass(caster, str)  # strings don't need to be cast
            and caster is not list  # annotation is list to begin with
        ):
            if caster in {tuple, deque}:
                pass
            elif _is_namedtuple_type(annotation):

                def _unpack(x):
                    return annotation(*x)  # type: ignore

                caster = _unpack
            else:  # pragma: no cover
                # covered in tests but coverage-tooling struggles here
                continue

            if param.kind is not _KEYWORD_ONLY:
                list_caster_by_pos[n] = caster
            if param.kind is not _POSITIONAL_ONLY:
                list_caster_by_name[name] = caster

    min_pos = min(list_caster_by_pos) if list_caster_by_pos else 0

    if not list_caster_by_name and not list_caster_by_pos:
        # no annotations associated with sequences
        return cast(_T, obj)

    @wraps(obj)
    def wrapper(*args, **kwargs):
        if list_caster_by_pos and min_pos < len(args):
            # at least one positional argument needs to be cast
            args = list(args)
            for pos in list_caster_by_pos:
                if pos < len(args):
                    if isinstance(args[pos], (list, ListConfig)):
                        args[pos] = list_caster_by_pos[pos](args[pos])
                else:
                    # pos is inserted in dict in ascending order
                    break
        if list_caster_by_name and kwargs:
            named = set(list_caster_by_name).intersection(kwargs)
            for name in named:
                # at least one named-arg needs to be cast
                if isinstance(kwargs[name], (list, ListConfig)):
                    kwargs[name] = list_caster_by_name[name](kwargs[name])

        return obj(*args, **kwargs)

    return cast(_T, wrapper)
